@@ -254,7 +254,11 @@ set_buffer_lines(
 		    && wp->w_cursor.lnum > append_lnum)
 		wp->w_cursor.lnum += added;
 	check_cursor_col();
-	update_topline();
+
+	// Only update the window view if w_buffer matches curbuf, otherwise
+	// the computations will be wrong.
+	if (curwin->w_buffer == curbuf)
+	    update_topline();
     }
 
 done:
@@ -567,10 +571,13 @@ f_deletebufline(typval_T *argvars, typval_T *rettv)
 	    {
 		if (wp->w_cursor.lnum > last)
 		    wp->w_cursor.lnum -= count;
-		else if (wp->w_cursor.lnum> first)
+		else if (wp->w_cursor.lnum > first)
 		    wp->w_cursor.lnum = first;
 		if (wp->w_cursor.lnum > wp->w_buffer->b_ml.ml_line_count)
 		    wp->w_cursor.lnum = wp->w_buffer->b_ml.ml_line_count;
+		wp->w_valid = 0;
+		if (wp->w_cursor.lnum <= wp->w_topline)
+		    wp->w_topline = 1;
 	    }
 	check_cursor_col();
 	deleted_lines_mark(first, count);
